// Copyright (c) 1997
// Michael A. Benzinger
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  The authors make no representation 
// about the suitability of this software for any purpose.  It is 
// provided "as is" without express or implied warranty.


// Disable warnings about the limit of 256 characters for identifiers imposed
// by the Microsoft Visual C++ debugger.
#ifdef _MSC_VER
  #pragma warning( disable : 4786 )
#endif

// Standard include files
#include <iostream>
#include <string>
#include <algorithm>

// Hash include files
#include "hashmap"
#include "hashmmap"
#include "hashset"
#include "hashmset"

// Promote the std namespace to the global namespace.
using namespace std;

// Define hash function for integers.  The hashing function is invoked through the
// () operator overload.
class int_hash_function
{
  public:
    int_hash_function(long primeArg = 1073741827) : prime(primeArg) {}
    ~int_hash_function() {}

    long
    operator()(const int& key) const
    {
      return( (key >= 0) ? (key % prime) : (-key % prime) );
    }

  protected:
    long prime;
};

// Define hash function for strings.  The hashing function is invoked through the
// () operator overload.
class str_hash_function
{
  public:
    str_hash_function(long primeArg = 1073741827) : prime(primeArg)  {}
    ~str_hash_function() {}

    long
    operator()(const string& str) const
    {
      int n = str.length();
      int h = 0; 
      
      const char* d = str.data();

      for ( int i = 0; i < n; ++i, ++d )
          h = (h << 2) + *d;

      return( (h >= 0) ? (h % prime) : (-h % prime) ); 
    }

  protected:
    long prime;
};


// Test the hash map container.
void
testHashMap()
{
  cout << "If the maps were constucted correctly, each of the two line pairs should match:" << endl;
  cout << endl;

  // Define a map using an integer as a key and a pointer to an integer as its value.
  cout << "Building a map with an integer key and a pointer to an integer" << endl;
  cout << "Forward Iterator" << endl;

  hash_map<int,int*,int_hash_function> hashIPtrMap;

  // Insert the test data into the map.
  hashIPtrMap[0] = new int(9);
  hashIPtrMap[1] = new int(8);
  hashIPtrMap[2] = new int(7);
  hashIPtrMap[3] = new int(6);
  hashIPtrMap[4] = new int(5);
  hashIPtrMap[5] = new int(4);
  hashIPtrMap[6] = new int(3);
  hashIPtrMap[7] = new int(2);
  hashIPtrMap[8] = new int(1);
  hashIPtrMap[9] = new int(0);

  // Define an iterator and iterate over all elements in the map.
  cout << "1.  {0,9},{1,8},{2,7},{3,6},{4,5},{5,4},{6,3},{7,2},{8,1},{9,0}" << endl;
  cout << "2.  ";

  hash_map<int,int*,int_hash_function>::iterator iptrItr;

  for ( iptrItr = hashIPtrMap.begin(); iptrItr != hashIPtrMap.end(); iptrItr++ )
  {
    if ( iptrItr != hashIPtrMap.begin() )
      cout << ",";

    cout << "{" << iptrItr->first << "," << *iptrItr->second << "}";
  }

  cout << endl;
  cout << endl;

  // Define a map using an integer as a key and a pointer to a string as its value.
  cout << "Building a map with an integer key and a pointer to a string" << endl;
  cout << "Forward Iterator" << endl;

  hash_map<int,string*,int_hash_function> hashSPtrMap;

  // Insert the test data into the map.
  hashSPtrMap[0] = new string("Five" );
  hashSPtrMap[1] = new string("Four" );
  hashSPtrMap[2] = new string("Three");
  hashSPtrMap[3] = new string("Two"  );
  hashSPtrMap[4] = new string("One"  );

  // Define an iterator and iterate over all elements in the map.
  cout << "1.  {0,Five},{1,Four},{2,Three},{3,Two},{4,One}" << endl;
  cout << "2.  ";

  hash_map<int,string*,int_hash_function>::iterator sptrItr;

  for ( sptrItr = hashSPtrMap.begin(); sptrItr != hashSPtrMap.end(); sptrItr++ )
  {
    if ( sptrItr != hashSPtrMap.begin() )
      cout << ",";

    cout << "{" << sptrItr->first << "," << *sptrItr->second << "}";
  }

  cout << endl;
  cout << endl;

  // Define a map using a string as a key and an integer as its value.
  cout << "Building a map with a string as a key and an integer" << endl;
  cout << "Forward and Reverse Iterators" << endl;

  hash_map<string, int, str_hash_function> hashStrMap;

  // Insert the test data into the map.
  hashStrMap["One"  ] = 1;
  hashStrMap["Two"  ] = 2;
  hashStrMap["Three"] = 3;
  hashStrMap["Four" ] = 4;
  hashStrMap["Five" ] = 5;

  // Define an iterator and iterate over all elements in the map.
  cout << "1.  {One,1},{Two,2},{Three,3},{Four,4},{Five,5}" << endl;
  cout << "2.  ";

  hash_map<string, int, str_hash_function>::iterator strItr;

  for ( strItr = hashStrMap.begin(); strItr != hashStrMap.end(); strItr++ )
  {
    if ( strItr != hashStrMap.begin() )
      cout << ",";

    cout << "{" << strItr->first << "," << strItr->second << "}";
  }

  cout << endl;
  cout << endl;

  // Define a reverse iterator and iterate over all elements in the map.
  hash_map<string, int, str_hash_function>::reverse_iterator rstrItr;

  cout << "3.  {Five,5},{Four,4},{Three,3},{Two,2},{One,1}" << endl;
  cout << "4.  ";

  for ( rstrItr = hashStrMap.rbegin(); rstrItr != hashStrMap.rend(); rstrItr++ )
  {
    if ( rstrItr != hashStrMap.rbegin() )
      cout << ",";

    cout << "{" << rstrItr->first << "," << rstrItr->second << "}";
  }

  cout << endl;
  cout << endl;

  // Test the STL search functions.  Start with the search function which requires two
  // sets of iterators.  One for the hash map to be searched and another subsequence
  // that is to be sought.
  hash_map<string, int, str_hash_function>::iterator firstStrItr;
  hash_map<string, int, str_hash_function>::iterator lastStrItr;

  firstStrItr = hashStrMap.find("Three");
  lastStrItr  = firstStrItr;

  lastStrItr++;

  strItr = search(hashStrMap.begin(), hashStrMap.end(), firstStrItr, lastStrItr);

  if ( strItr != hashStrMap.end() )
    cout << "search(): pair<\"Three\",3>={" << strItr->first << "," << strItr->second << "}" << endl;
  else
    cout << "Search(): pair<\"Three\",3>={}  Function failed!!!" << endl;

  // Test search_n to find a given value.
  strItr = search_n(hashStrMap.begin(), hashStrMap.end(), 1, pair<string,int>("Four",4));

  if ( strItr != hashStrMap.end() )
    cout << "search_n(): pair<\"Four\",4>={" << strItr->first << "," << strItr->second << "}" << endl;
  else
    cout << "search_n(): pair<\"Four\",4>={}  Function failed!!!" << endl;

  cout << endl;

  // Create a second hash map that is built identically to the previous one.
  cout << "Building a second map with a string as a key and an integer" << endl;
  cout << "Forward Iterator and Comparisons" << endl;

  hash_map<string, int, str_hash_function> hashStrMap2;

  // Insert the test data into the map.
  hashStrMap2["One"  ] = 1;
  hashStrMap2["Two"  ] = 2;
  hashStrMap2["Three"] = 3;
  hashStrMap2["Four" ] = 4;
  hashStrMap2["Five" ] = 5;

  // Iterate over all elements in the map to demonstrate equality.
  cout << "1.  {One,1},{Two,2},{Three,3},{Four,4},{Five,5}" << endl;
  cout << "2.  ";

  for ( strItr = hashStrMap2.begin(); strItr != hashStrMap2.end(); strItr++ )
  {
    if ( strItr != hashStrMap2.begin() )
      cout << ",";

    cout << "{" << strItr->first << "," << strItr->second << "}";
  }

  cout << endl;
  cout << endl;

  // Compare the two hash tables for equality.
  cout << "Equality comparison:  These two hash maps should compare equal" << endl;

  if ( hashStrMap == hashStrMap2 )
    cout << "  hashStrMap == hashStrMap2" << endl;
  else
    cout << "  hashStrMap != hashStrMap2" << endl;

  // Change the value for a greater than comparison.
  cout << endl;
  cout << "Change value for hashStrMap2 key \"Five\" to zero" << endl;
  cout << endl;

  hashStrMap2["Five"] = 0;

  // Compare the two hash tables for equality.
  cout << "Greater than comparison:  hashStrMap should compare greater than hashStrMap2" << endl;

  if ( hashStrMap > hashStrMap2 )
    cout << "  hashStrMap > hashStrMap2" << endl;
  else
    cout << "  hashStrMap <= hashStrMap2" << endl;

  cout << endl;

  // Define a map using an integer as a key and an integer as its value.
  cout << "Building a map with an integer key and an integer" << endl;
  cout << "Insert using operator[]:  Value for hashMap[20]=15" << endl;

  hash_map<int, int, int_hash_function> hashMap;

  // Insert values into the hash map using the operator[].  Output the value
  // with the key 20 as a test.
  hashMap[20] = 15;
  hashMap[21] = 16;
  hashMap[22] = 17;
  hashMap[23] = 18;

  cout << "  hashMap[20]=" << hashMap[20] << endl;

  // Define a pair with a hash map iterator and a boolean value returned
  // when a value is inserted.
  pair<hash_map<int, int, int_hash_function>::iterator, bool> posItr;

  // Attempt to insert {5,22} for the first time and report success or failure.
  posItr = hashMap.insert(pair<int,int>(5, 22));

  if ( posItr.second )
    cout << "  First try: {5,22} Inserted" << endl;
  else
    cout << "  First try: {5,22} NOT Inserted" << endl;

  // Attempt to insert {5,22} for the second time and report success or failure.
  posItr = hashMap.insert(pair<int,int>(5, 22));

  if ( posItr.second )
    cout << "  Second try: {5,22} Inserted" << endl;
  else
    cout << "  Second try: {5,22} NOT Inserted" << endl;

  // Attempt to insert {6,29} for the first time and report success or failure.
  posItr = hashMap.insert(pair<int,int>(6, 29));

  if ( posItr.second )
    cout << "  First try: {6,29} Inserted" << endl;
  else
    cout << "  First try: {6,29} NOT Inserted" << endl;

  // If the value was inserted, then delete it.
  if ( posItr.second )
    hashMap.erase(posItr.first);

  // Output the counts of the given values.
  cout << "  Count of 5 should be 1 and count of 6 should be 0: " << endl;
  cout << "    count(5)=" << hashMap.count(5) << endl;
  cout << "    count(6)=" << hashMap.count(6) << endl;

  // Insert three map entries from an array of key/value pairs and erase one once
  // it has been inserted.
  pair<int,int> anArray[3] =
    {
      pair<int,int>(156, -1),
      pair<int,int>(157, -2),
      pair<int,int>(158, -3)
    };

  cout << "  Inserting {156,-1},{157,-2},{158,-3}" << endl;
  cout << "  Erasing {156,-1}" << endl;

  hashMap.insert(anArray, &anArray[3]);

  hashMap.erase(156);

  // Declare two iterators and assign the first to point to the key 21
  // and the last to point to key 23.  Invoke erase on the iterators to
  // delete  keys 21 and 22.
  cout << "  Erasing {21,16} and {22,17}" << endl;

  hash_map<int, int, int_hash_function>::iterator first;
  hash_map<int, int, int_hash_function>::iterator last;

  first = hashMap.find(21);
  last  = hashMap.find(23);

  hashMap.erase(first, last);

  // Define a new hash map that will be a copy and swap it with the hash map with
  // which we have been working.
  cout << "  Swapping contents of hashMap with hashMapCopy" << endl;

  hash_map<int, int, int_hash_function> hashMapCopy;

  hashMapCopy.swap(hashMap);

  // Define an iterator and output the contents of the first hash map.
  cout << "  hashMap: Forward iterator results should indicate an empty hash map" << endl;
  cout << "    1." << endl;
  cout << "    2." << endl;

  hash_map<int, int, int_hash_function>::iterator mapItr;

  for ( mapItr = hashMap.begin(); mapItr != hashMap.end(); mapItr++ )
    cout << "  " << (*mapItr).first << "  " << (*mapItr).second << endl;

  cout << "  hashMapCopy: Forward iterator should contain the previous contents of hashMap" << endl;
  cout << "    1.  {20,15},{23,18},{5,22},{157,-2},{158,-3}" << endl;
  cout << "    2.  ";

  for ( mapItr = hashMapCopy.begin(); mapItr != hashMapCopy.end(); mapItr++ )
  {
    if ( mapItr != hashMapCopy.begin() )
      cout << ",";

    cout << "{" << (*mapItr).first << "," << (*mapItr).second << "}";
  }

  cout << endl;
  cout << endl;

  // Compare hashMap and hashMapCopy using the equality and less than operators.
  cout << "  hashMap should not be equal to hashMapCopy" << endl;

  if ( hashMap == hashMapCopy )
    cout << "    hashMap == hashMapCopy" << endl;
  else
    cout << "    hashMap != hashMapCopy" << endl;

  cout << "  hashMap should be less than hashMapCopy" << endl;

  if ( hashMap < hashMapCopy )
    cout << "    hashMap < hashMapCopy" << endl;
  else
    cout << "    hashMap >= hashMapCopy" << endl;

  cout << endl;

  // Make a copy of hashMapCopy using the copy constructor.
  hash_map<int, int, int_hash_function> hashMapCopy2(hashMapCopy);

  cout << "hashMapCopy2: Forward iterator should contain a copy of the contents of hashMapCopy" << endl;
  cout << "    1.  {20,15},{23,18},{5,22},{157,-2},{158,-3}" << endl;
  cout << "    2.  ";

  for ( mapItr = hashMapCopy2.begin(); mapItr != hashMapCopy2.end(); mapItr++ )
  {
    if ( mapItr != hashMapCopy2.begin() )
      cout << ",";

    cout << "{" << (*mapItr).first << "," << (*mapItr).second << "}";
  }

  cout << endl;
  cout << endl;

  // Compare the original hash map and its copy.
  cout << "  hashMapCopy should be equal to hashMapCopy2" << endl;

  if ( hashMapCopy == hashMapCopy2 )
    cout << "    hashMapCopy == hashMapCopy2" << endl;
  else
    cout << "    hashMapCopy != hashMapCopy2" << endl;

  cout << endl;

  // Resize the hashMapCopy and output its contents.
  cout << "Resize hashMapCopy to 51 buckets" << endl;
  cout << endl;

  hashMapCopy.resize(51);

  // Output the contents of the resized hash map.
  cout << "hashMapCopy resized: Forward" << endl;
  cout << "    1.  {20,15},{23,18},{158,-3},{5,22},{157,-2}" << endl;
  cout << "    2.  ";

  for ( mapItr = hashMapCopy.begin(); mapItr != hashMapCopy.end(); mapItr++ )
  {
    if ( mapItr != hashMapCopy.begin() )
      cout << ",";

    cout << "{" << (*mapItr).first << "," << (*mapItr).second << "}";
  }

  cout << endl;
  cout << endl;

  // Define a reverse iterator and output the resized hash map copy in reverse order.
  hash_map<int, int, int_hash_function>::reverse_iterator rintItr;

  cout << "hashMapCopy resized: Reverse" << endl;
  cout << "    1.  {157,-2},{5,22},{158,-3},{23,18},{20,15}" << endl;
  cout << "    2.  ";

  for ( rintItr = hashMapCopy.rbegin(); rintItr != hashMapCopy.rend(); rintItr++ )
  {
    if ( rintItr != hashMapCopy.rbegin() )
      cout << ",";

    cout << "{" << (*rintItr).first << "," << (*rintItr).second << "}";
  }

  cout << endl;
  cout << endl;

  // Define a new hash map and insert values that are equal to those in hashMapCopy.
  // These will be used to test the comparison operators.
  hash_map<int, int, int_hash_function> hashMap2;

  hashMap2[  5] = 22;
  hashMap2[ 23] = 18;
  hashMap2[ 20] = 15;
  hashMap2[157] = -2;
  hashMap2[158] = -3;

  cout << "Building a hash map with an integer key and an integer" << endl;
  cout << "Forward Iterator and Comparisons" << endl;

  // Output the contents of the hash map.
  cout << "hashMap2 " << endl;
  cout << "    1.  {5,22},{23,18},{20,15},{157,-2},{158,-3}" << endl;
  cout << "    2.  ";

  for ( mapItr = hashMap2.begin(); mapItr != hashMap2.end(); mapItr++ )
  {
    if ( mapItr != hashMapCopy.begin() )
      cout << ",";

    cout << "{" << (*mapItr).first << "," << (*mapItr).second << "}";
  }

  cout << endl;
  cout << endl;

  cout << "  hashMap2 should compare equal to hashMapCopy2" << endl;

  if ( hashMap2 == hashMapCopy2 )
    cout << "    hashMap2 == hashMapCopy2" << endl;
  else
    cout << "    hashMap2 != hashMapCopy2" << endl;

  cout << "  hashMap2 should compare less than or equal to hashMapCopy2" << endl;

  if ( hashMap2 <= hashMapCopy2 )
    cout << "    hashMap2 <= hashMapCopy2" << endl;
  else
    cout << "    hashMap2 >  hashMapCopy2" << endl;

  cout << "  Modify key 5 to 23 and key 23 to 19 in hashMap2." << endl;
  cout << "  HashMap2 should then compare greater than or equal to hashMapCopy2" << endl;

  hashMap2[  5] = 23;
  hashMap2[ 23] = 19;

  if ( hashMap2 < hashMapCopy2 )
    cout << "    hashMapCopy < hashMapCopy2" << endl;
  else
    cout << "    hashMapCopy >= hashMapCopy2" << endl;

  cout << endl;
}


// Test the hash multimap container.
void
testHashMultiMap()
{
  // Build a hash multimap with an integer as a key and an integer as a value.
  cout << "Building a hash multimap with an integer key and an integer" << endl;
  cout << "Forward and Reverse Iterators" << endl;

  hash_multimap<int, int, int_hash_function> hashMultiMap;

  pair<hash_multimap<int, int, int_hash_function>::iterator, bool> mposItr;

  mposItr = hashMultiMap.insert(pair<int,int>(5, 22));
  mposItr = hashMultiMap.insert(pair<int,int>(6, 29));
  mposItr = hashMultiMap.insert(pair<int,int>(6, 30));
  mposItr = hashMultiMap.insert(pair<int,int>(6, 31));
  mposItr = hashMultiMap.insert(pair<int,int>(6, 32));
  mposItr = hashMultiMap.insert(pair<int,int>(6, 34));
  mposItr = hashMultiMap.insert(pair<int,int>(7, 35));

  // Output the count2 of the given values.
  cout << "  Count of 5 should be 1 and count of 6 should be 5: " << endl;
  cout << "    count(5)=" << hashMultiMap.count(5) << endl;
  cout << "    count(6)=" << hashMultiMap.count(6) << endl;
  cout << endl;

  // Output the contents of the hash map with a forward iterator.
  cout << "hashMultiMap: Forward Iterator" << endl;
  cout << "    1.  {5,22},{6,29},{6,30},{6,31},{6,32},{6,34},{7,35}" << endl;
  cout << "    2.  ";

  hash_multimap<int, int, int_hash_function>::iterator mapItr;

  for ( mapItr = hashMultiMap.begin(); mapItr != hashMultiMap.end(); mapItr++ )
  {
    if ( mapItr != hashMultiMap.begin() )
      cout << ",";

    cout << "{" << (*mapItr).first << "," << (*mapItr).second << "}";
  }

  cout << endl;
  cout << endl;

  // Output the contents of the hash map with a forward iterator.
  cout << "hashMultiMap: Reverse Iterator" << endl;
  cout << "    1.  {7,35},{6,34},{6,33},{6,32},{6,30},{6,29},{5,22}" << endl;
  cout << "    2.  ";

  hash_multimap<int, int, int_hash_function>::reverse_iterator rmapItr;

  for ( rmapItr = hashMultiMap.rbegin(); rmapItr != hashMultiMap.rend(); rmapItr++ )
  {
    if ( rmapItr != hashMultiMap.rbegin() )
      cout << ",";

    cout << "{" << (*rmapItr).first << "," << (*rmapItr).second << "}";
  }

  cout << endl;
  cout << endl;
}


// Test the hash set container.
void
testHashSet()
{
  // Build a hash set with an integer as a value.
  cout << "Building a hash set with an integer value" << endl;
  cout << "Forward and Reverse Iterators" << endl;

  hash_set<int, int_hash_function, equal_to<int> > hashSet;

  hashSet.insert(11);
  hashSet.insert(12);
  hashSet.insert(13);
  hashSet.insert(14);
  hashSet.insert(14);
  hashSet.insert(15);

  // Output the contents of the hash set with a forward iterator.
  cout << "hashSet: Forward Iterator" << endl;
  cout << "    1.  11,12,13,14,15" << endl;
  cout << "    2.  ";

  hash_set<int, int_hash_function>::iterator setItr;

  for ( setItr = hashSet.begin(); setItr != hashSet.end(); setItr++ )
  {
    if ( setItr != hashSet.begin() )
      cout << ",";

    cout << *setItr;
  }

  cout << endl;
  cout << endl;

  // Output the contents of the hash set with a reverse iterator.
  cout << "hashSet: Reverse Iterator" << endl;
  cout << "    1.  15,14,13,12,11" << endl;
  cout << "    2.  ";

  hash_set<int, int_hash_function>::reverse_iterator rsetItr;

  for ( rsetItr = hashSet.rbegin(); rsetItr != hashSet.rend(); rsetItr++ )
  {
    if ( rsetItr != hashSet.rbegin() )
      cout << ",";

    cout << *rsetItr;
  }

  cout << endl;
  cout << endl;
}


// Test the hash multiset container.
void
testHashMultiSet()
{
  // Build a hash set with an integer as a value.
  cout << "Building a hash multiset with an integer value" << endl;
  cout << "Forward and Reverse Iterators" << endl;

  hash_multiset<int, int_hash_function> hashMultiSet;

  hashMultiSet.insert(115);
  hashMultiSet.insert(116);
  hashMultiSet.insert(116);
  hashMultiSet.insert(116);
  hashMultiSet.insert(117);

  cout << "  Count of 116 should be 3" << endl;
  cout << "    count(116)=" << hashMultiSet.count(116) << endl;
  cout << endl;

  // Output the contents of the hash multiset with a forward iterator.
  cout << "hashMultiSet: Forward Iterator" << endl;
  cout << "    1.  115,116,116,116,117" << endl;
  cout << "    2.  ";

  hash_multiset<int, int_hash_function>::iterator setItr;

  for ( setItr = hashMultiSet.begin(); setItr != hashMultiSet.end(); setItr++ )
  {
    if ( setItr != hashMultiSet.begin() )
      cout << ",";

    cout << *setItr;
  }

  cout << endl;
  cout << endl;

  // Output the contents of the hash set with a reverse iterator.
  cout << "hashMultiSet: Reverse Iterator" << endl;
  cout << "    1.  117,116,116,116,115" << endl;
  cout << "    2.  ";

  hash_set<int, int_hash_function>::reverse_iterator rsetItr;

  for ( rsetItr = hashMultiSet.rbegin(); rsetItr != hashMultiSet.rend(); rsetItr++ )
  {
    if ( rsetItr != hashMultiSet.rbegin() )
      cout << ",";

    cout << *rsetItr;
  }

  cout << endl;
  cout << endl;
}

// Define the main routine for testing the four hash containgers: map, multimap, set and mutliset.
int
main(int argc, char* argv[])
{
  testHashMap();

  cout << "-----------------------------------------------------------------" << endl;
  cout << endl;

  testHashMultiMap();

  cout << "-----------------------------------------------------------------" << endl;
  cout << endl;

  testHashSet();

  cout << "-----------------------------------------------------------------" << endl;
  cout << endl;

  testHashMultiSet();

  return( 0 );
}
